"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports._TSHashBuffer = void 0;
const commons_1 = require("./commons");
const data_1 = require("./data");
const tsdata_1 = require("./tsdata");
const tserrors_1 = require("./tserrors");
const utils_1 = require("./utils");
class _TSHashBuffer {
    constructor(buf, opts) {
        if (!(0, commons_1.$defined)(opts.padding)) {
            opts.padding = [0x80];
        }
        this.sourceLength = (0, commons_1.$length)(buf);
        this.paddingLength = (0, commons_1.$length)(opts.padding);
        let blen = this.sourceLength + this.paddingLength;
        blen = blen + (blen % 4 > 0 ? 4 : 0);
        this.intCount = opts.integersCount;
        this.blockSize = opts.integersCount * 4;
        if (this.paddingLength > Math.floor(this.blockSize / 8)) {
            throw new tserrors_1.TSError('Impossible to create HashBlock with a padding this size', {
                sourceLength: this.sourceLength,
                paddingLength: this.paddingLength,
                blockSize: this.blockSize,
            });
        }
        this.blocksCount = Math.ceil((blen + 8) / this.blockSize);
        this._block = Buffer.alloc(this.blockSize);
        this._ret = new Array(this.intCount);
        this._endianness = (0, commons_1.$value)(opts.endianness, 'BE');
        if (this.sourceLength > 0) {
            if (buf instanceof tsdata_1.TSData) {
                const [store,] = buf.internalStorage;
                this._source = store;
            }
            else if (buf instanceof ArrayBuffer) {
                this._source = (0, data_1.$uint8ArrayFromArrayBuffer)(buf);
            }
            else {
                this._source = buf;
            }
        }
        if (this.paddingLength > 0) {
            this._padding = opts.padding;
        }
    }
    _getBlock(blockIndex, debug) {
        if (blockIndex >= this.blocksCount) {
            throw '_TSHashBuffer._getBlock index overflow';
        }
        let i = blockIndex * this.blockSize, j = 0;
        const endBlock = i + this.blockSize;
        const end = Math.min(endBlock, this.sourceLength);
        if (debug) {
            (0, utils_1.$logterm)(`&0\n&xi = &y${i}&x, end = &y${end}&x, endBlock = &y${endBlock}&0`);
        }
        for (; i < end; i++) {
            this._block[j++] = this._source[i];
        }
        if (i < endBlock) {
            if (i <= end) {
                for (let p = 0; p < this.paddingLength; p++) {
                    this._block[j++] = (this._padding)[p];
                }
            }
            while (j < this.blockSize) {
                this._block[j++] = 0x00;
            }
            if (blockIndex + 1 === this.blocksCount) {
                const bits = this.sourceLength * 8;
                const lowBits = (bits & 0xffffffff) >>> 0;
                const highBits = (((bits - lowBits) / 0x100000000) & 0xffffffff) >>> 0;
                if (this._endianness === 'BE') {
                    if (highBits) {
                        this._block.writeUInt32BE(highBits, this.blockSize - 8);
                    }
                    if (lowBits) {
                        this._block.writeUInt32BE(lowBits, this.blockSize - 4);
                    }
                }
                else {
                    if (lowBits) {
                        this._block.writeUInt32LE(lowBits, this.blockSize - 8);
                    }
                    if (highBits) {
                        this._block.writeUInt32LE(highBits, this.blockSize - 4);
                    }
                }
            }
        }
        if (this._endianness === 'BE') {
            for (let n = 0, offset = 0; n < this.intCount; n++, offset += 4) {
                this._ret[n] = this._block.readUInt32BE(offset);
            }
        }
        else {
            for (let n = 0, offset = 0; n < this.intCount; n++, offset += 4) {
                this._ret[n] = this._block.readUInt32LE(offset);
            }
        }
        if (debug) {
            (0, utils_1.$logterm)(`&0&cblock[${blockIndex}]:&0`);
            (0, utils_1.$hexadump)(this._block);
        }
        return this._ret;
    }
    hash(fn, printBlocks) {
        const debug = !!printBlocks;
        for (let i = 0; i < this.blocksCount; i++) {
            fn(this._getBlock(i, debug));
        }
    }
    output(h) {
        const len = h.length;
        const buf = Buffer.allocUnsafe(len * 4);
        if (this._endianness === 'BE') {
            for (let i = 0, offset = 0; i < len; i++, offset += 4) {
                buf.writeUInt32BE(h[i], offset);
            }
        }
        else {
            for (let i = 0, offset = 0; i < len; i++, offset += 4) {
                buf.writeUInt32LE(h[i], offset);
            }
        }
        return !this.dataoutput ? buf.toString('hex') : buf;
    }
}
exports._TSHashBuffer = _TSHashBuffer;
//# sourceMappingURL=crypto_hashBuffer.js.map