"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSSystemCharset = exports.TSCharset = exports.$charset = void 0;
const commons_1 = require("./commons");
const data_1 = require("./data");
const string_tables_1 = require("./string_tables");
const tscharsets_json_1 = __importDefault(require("./tscharsets.json"));
const tsdata_1 = require("./tsdata");
const tserrors_1 = require("./tserrors");
const types_1 = require("./types");
const utils_1 = require("./utils");
function $charset(value, defaultCharset = TSCharset.utf8Charset()) { return (0, commons_1.$isstring)(value) ? TSCharset.encoding(value) : (0, commons_1.$value)(value, defaultCharset); }
exports.$charset = $charset;
var TSCachedCharset;
(function (TSCachedCharset) {
    TSCachedCharset[TSCachedCharset["ASCII"] = 0] = "ASCII";
    TSCachedCharset[TSCachedCharset["Latin1"] = 1] = "Latin1";
    TSCachedCharset[TSCachedCharset["UTF8"] = 2] = "UTF8";
    TSCachedCharset[TSCachedCharset["UTF16"] = 3] = "UTF16";
    TSCachedCharset[TSCachedCharset["ANSI"] = 4] = "ANSI";
    TSCachedCharset[TSCachedCharset["MAC"] = 5] = "MAC";
})(TSCachedCharset || (TSCachedCharset = {}));
;
class TSCharset {
    static asciiCharset() { return TSCharset._cachedCharset(TSCachedCharset.ASCII, 'ascii'); }
    static latin1Charset() { return TSCharset._cachedCharset(TSCachedCharset.Latin1, 'latin1'); }
    static binaryCharset() { return TSCharset.latin1Charset(); }
    static utf8Charset() { return TSCharset._cachedCharset(TSCachedCharset.UTF8, 'utf8'); }
    static unicodeCharset() { return TSCharset._cachedCharset(TSCachedCharset.UTF16, 'utf16le'); }
    static ansiCharset() { return TSCharset._cachedCharset(TSCachedCharset.UTF16, 'ansi'); }
    static macCharset() { return TSCharset._cachedCharset(TSCachedCharset.UTF16, 'mac'); }
    constructor(name, aliases) {
        if (!(0, commons_1.$length)(name)) {
            throw new tserrors_1.TSError('TSCharset.constructor(): trying to instanciate unamed RTF charset', { aliases: aliases });
        }
        this.name = name;
        this.aliases = (0, commons_1.$ok)(aliases) ? [...aliases] : [];
    }
    static systemCharset() {
        return (0, commons_1.$value)(TSCharset.charset(_systemEncoding()), TSCharset.utf8Charset());
    }
    static encoding(enc) { return this.charset((0, commons_1.$encoding)(enc)); }
    static charset(name) {
        if (!(0, commons_1.$ok)(TSCharset.__charsetsMap)) {
            TSCharset.__charsetsMap = new Map();
            tscharsets_json_1.default.forEach(def => {
                let charset = new TSLoadedCharset(def);
                TSCharset.__charsetsMap.set(def.name.toLowerCase(), charset);
                charset.aliases.forEach(a => {
                    TSCharset.__charsetsMap.set(a.toLowerCase(), charset);
                });
            });
            ['ANSI', 'MacRoman'].forEach(c => {
                if (!(0, commons_1.$ok)(TSCharset.__charsetsMap.get(c.toLowerCase()))) {
                    throw new tserrors_1.TSError(`TSCharset.charset() : system loaded charset '${c}' is missing.`, { name: name });
                }
            });
            string_tables_1.FoundationEncodingsAliases.forEach(def => {
                let set = new Set;
                const name = def.name.toLowerCase();
                set.add(name);
                def.aliases.forEach(a => set.add(a.toLowerCase()));
                set.delete(name);
                const aliases = Array.from(set);
                let charset = new TSSystemCharset(name, aliases);
                TSCharset.__charsetsMap.set(name, charset);
                aliases.forEach(a => TSCharset.__charsetsMap.set(a, charset));
            });
        }
        return TSCharset.__charsetsMap.get(name.toLowerCase());
    }
    static _cachedCharset(code, name) {
        if (!(0, commons_1.$ok)(TSCharset.__cachedCharsets[code])) {
            TSCharset.__cachedCharsets[code] = TSCharset.charset(name);
        }
        return TSCharset.__cachedCharsets[code];
    }
    stringFromData(source, sourceStart, sourceEnd) {
        return this.stringFromBytes((0, data_1.$bytesFromDataLike)(source), sourceStart, sourceEnd);
    }
    bufferFromString(source, sourceStart, sourceEnd) {
        return (0, data_1.$bufferFromBytes)(this.stringToBytes(source, sourceStart, sourceEnd));
    }
    uint8ArrayFromString(source, sourceStart, sourceEnd) {
        return (0, data_1.$uint8ArrayFromBytes)(this.stringToBytes(source, sourceStart, sourceEnd));
    }
    bytesFromString(source, sourceStart, sourceEnd) {
        return (0, data_1.$arrayFromBytes)(this.stringToBytes(source, sourceStart, sourceEnd));
    }
    dataFromString(source, sourceStart, sourceEnd) {
        return new tsdata_1.TSData(this.bufferFromString(source), { dontCopySourceBuffer: true });
    }
}
exports.TSCharset = TSCharset;
TSCharset.__charsetsMap = undefined;
TSCharset.__cachedCharsets = [];
class TSLoadedCharset extends TSCharset {
    constructor(def) {
        const n = TSLoadedCharset.UnicodeBufferSize;
        super(def.name, def.aliases);
        if ((0, commons_1.$count)(def.charset) !== 256) {
            throw new tserrors_1.TSError(`TSLoadedCharset.constructor(): charset ${def.name} has an invalid charset count (must be 256)`, { definition: def });
        }
        this._toUnicodeTable = [];
        this._fromUnicodeTable = new Uint8Array(n);
        this._fromUnicodeMap = new Map();
        for (let i = 0; i < 256; i++) {
            const uc = (0, commons_1.$toint)(def.charset[i]);
            if (uc < -1 || uc > types_1.UINT16_MAX) {
                throw new tserrors_1.TSError(`TSLoadedCharset.constructor(): charset '${def.name}': Bad unicode 16 character ar charset[${i}]`, { definition: def });
            }
            if (uc >= 0) {
                if (uc < n) {
                    this._fromUnicodeTable[uc] = i;
                }
                else {
                    this._fromUnicodeMap.set(uc, i);
                }
                this._toUnicodeTable[i] = String.fromCharCode(uc);
            }
        }
    }
    stringToBytes(source, sourceStart, sourceEnd) {
        const [, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        const n = TSLoadedCharset.UnicodeBufferSize;
        const ret = [];
        for (let i = start; i < end; i++) {
            const uc = source.charCodeAt(i);
            const c = uc < n ? this._fromUnicodeTable[uc] : this._fromUnicodeMap.get(uc);
            if ((0, commons_1.$ok)(c)) {
                ret.push(c);
            }
        }
        return ret;
    }
    stringFromBytes(source, sourceStart, sourceEnd) {
        const [, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        let target = '';
        if (start < end) {
            for (let i = start; i < end; i++) {
                const s = this._toUnicodeTable[source[i]];
                if (s) {
                    target += s;
                }
            }
        }
        return target;
    }
}
TSLoadedCharset.UnicodeBufferSize = 256;
class TSSystemCharset extends TSCharset {
    stringToBytes(source, sourceStart, sourceEnd) {
        const [len, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        if (start < end) {
            const s = start === 0 && end === len ? source : source.slice(start, end);
            return Buffer.from(s, (0, commons_1.$encoding)(this.name));
        }
        return [];
    }
    _stringFromBuffer(source, sourceStart, sourceEnd) {
        const [, start, end,] = (0, commons_1.$lse)(source, sourceStart, sourceEnd);
        if (start < end) {
            return source.toString((0, commons_1.$encoding)(this.name), start, end);
        }
        return '';
    }
    stringFromBytes(source, sourceStart, sourceEnd) {
        return this._stringFromBuffer((0, data_1.$bufferFromBytes)(source), sourceStart, sourceEnd);
    }
    stringFromData(src, sourceStart, sourceEnd) {
        if (src instanceof Buffer) {
            return this._stringFromBuffer(src, sourceStart, sourceEnd);
        }
        else if (src instanceof tsdata_1.TSData) {
            return this._stringFromBuffer(src.mutableBuffer, sourceStart, sourceEnd);
        }
        else {
            return this._stringFromBuffer((0, data_1.$bufferFromBytes)(src), sourceStart, sourceEnd);
        }
    }
}
exports.TSSystemCharset = TSSystemCharset;
;
function _systemEncoding() {
    if ((0, utils_1.$inbrowser)()) {
        return document.characterSet;
    }
    const p = (0, commons_1.$isproperty)(process, 'platform') ? process === null || process === void 0 ? void 0 : process.platform : 'linux';
    switch (p) {
        case 'darwin': return 'MacRoman';
        case 'win32':
        case 'cygwin': return 'ANSI';
        default: return 'latin1';
    }
}
//# sourceMappingURL=tscharset.js.map