"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSDataTemplate = exports.TSHTMLTemplate = exports.TSGenericDataTemplate = exports.TSStringTemplate = exports.TSFusionTemplate = exports.TSCharRole = void 0;
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const data_1 = require("./data");
const tscharset_1 = require("./tscharset");
const tsdata_1 = require("./tsdata");
const tsdate_1 = require("./tsdate");
const tserrors_1 = require("./tserrors");
const tsfusionnode_1 = require("./tsfusionnode");
const utils_1 = require("./utils");
var TSCharRole;
(function (TSCharRole) {
    TSCharRole[TSCharRole["UnsignificantStarterChar"] = 0] = "UnsignificantStarterChar";
    TSCharRole[TSCharRole["StarterChar"] = 1] = "StarterChar";
    TSCharRole[TSCharRole["InternalChar"] = 2] = "InternalChar";
    TSCharRole[TSCharRole["ForbiddenChar"] = 3] = "ForbiddenChar";
})(TSCharRole = exports.TSCharRole || (exports.TSCharRole = {}));
;
const [TAB, LF, BS, FF, CR, SPACE, EXCLAMATION_MARK, DOUBLE_QUOTE, SHARP, DOLLAR, QUOTE, LPAR, RPAR, ASTERISK, PLUS, COMMA, MINUS, DOT, SLASH, ZERO, NINE, COLON, LT, EQ, GT, QUESTION_MARK, AT, A, Uu, Z, BACKSLASH, UNDERSCORE, bu, fu, nu, ru, tu, uu, PIPE] = [0x09, 0x0a, 0x0B, 0x0C, 0x0D, 0x20, 0x21, 0x22, 0x23, 0x24, 0x27, 0x28, 0x29, 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F, 0x30, 0x39, 0x3A, 0x3C, 0x3D, 0x3E, 0x3F, 0x40, 0x41, 0x55, 0x5A, 0x5C, 0x5F, 0x62, 0x66, 0x6E, 0x72, 0x74, 0x75, 0x7C];
const FUSION_SPECIAL_CHARS = [SHARP, QUESTION_MARK, EXCLAMATION_MARK, DOLLAR, AT, UNDERSCORE, ASTERISK, PLUS, DOT];
const TSFusionNodeTypes = [];
TSFusionNodeTypes[SHARP] = tsfusionnode_1.TSFusionNodeType.Enumeration;
TSFusionNodeTypes[QUESTION_MARK] = tsfusionnode_1.TSFusionNodeType.Test;
TSFusionNodeTypes[EXCLAMATION_MARK] = tsfusionnode_1.TSFusionNodeType.NegativeTest;
const TSFusionContextTypes = [];
TSFusionContextTypes[DOLLAR] = tsfusionnode_1.TSFusionContextType.Global;
TSFusionContextTypes[PLUS] = tsfusionnode_1.TSFusionContextType.User;
TSFusionContextTypes[AT] = tsfusionnode_1.TSFusionContextType.Root;
TSFusionContextTypes[UNDERSCORE] = tsfusionnode_1.TSFusionContextType.System;
TSFusionContextTypes[ASTERISK] = tsfusionnode_1.TSFusionContextType.Procedure;
TSFusionContextTypes[DOT] = tsfusionnode_1.TSFusionContextType.Local;
const TSFusionForbiddenEncodings = ['utf16le', 'base64', 'base64url', 'hex'];
class TSFusionTemplate {
    static fromData(source, encoding, opts) { return TSDataTemplate.fromData(source, encoding, opts); }
    static fromHTMLData(source, opts) { return TSHTMLTemplate.fromHTMLData(source, opts); }
    static fromString(source, opts) { return TSStringTemplate.fromString(source, opts); }
    constructor(src, opts = {}, paramCharset = tscharset_1.TSCharset.utf8Charset()) {
        this._capacity = 0;
        this._procedures = {};
        this._variables = new Map();
        this._contextsVariables = new Map();
        const sm = _ascii2data((0, commons_1.$length)(opts.startingMark) ? opts.startingMark : TSFusionTemplate.DefaultStartingMark);
        const em = _ascii2data((0, commons_1.$length)(opts.endingMark) ? opts.endingMark : TSFusionTemplate.DefaultEndingMark);
        const sepa = _ascii2data((0, commons_1.$length)(opts.separator) ? opts.separator : TSFusionTemplate.DefaultSeparator);
        const debug = !!opts.debugParsing;
        const source = (0, data_1.$uint8ArrayFromDataLike)(src);
        const constants = [
            { name: _ascii2data('FUTURE'), value: tsdate_1.TSDate.future(), len: 6 },
            { name: _ascii2data('NaN'), value: NaN, len: 3 },
            { name: _ascii2data('PAST'), value: tsdate_1.TSDate.past(), len: 4 },
            { name: _ascii2data('current'), value: this, len: 7 },
            { name: _ascii2data('false'), value: false, len: 4 },
            { name: _ascii2data('null'), value: null, len: 4 },
            { name: _ascii2data('true'), value: true, len: 4 },
            { name: _ascii2data('undefined'), value: undefined, len: 9 }
        ];
        if (debug) {
            (0, utils_1.$logterm)(`&0\n&Y&k ${this.constructor.name} parsing in debug mode &0`);
            (0, utils_1.$logterm)(`&0&x - &w starting mark&x:        '&o${_data2ascii(sm)}&x'&0`);
            (0, utils_1.$logterm)(`&0&x - &w ending mark&x:          '&o${_data2ascii(em)}&x'&0`);
            (0, utils_1.$logterm)(`&0&x - &w separator&x:            '&o${_data2ascii(sepa)}&x'&0`);
            (0, utils_1.$logterm)('&0&x======== &wsource&x: ========&0\n&y' + (0, utils_1.$inspect)(source) + "&0");
        }
        if ((0, compare_1.$equal)(sm, em) || (0, compare_1.$equal)(sepa, sm) || (0, compare_1.$equal)(sepa, em)) {
            throw new tserrors_1.TSError('starting mark, ending mark and seperator must be distinct', { startingMark: sm, endingMark: em, separator: sepa });
        }
        const len = source.length;
        const smlen = sm.length;
        const sepalen = sepa.length;
        const emlen = em.length;
        function upper(c) { return (c & 0xdf); }
        function isVariableSpace(c) { return c === TAB || c === SPACE; }
        function isVariableNum(c) { return c >= ZERO && c <= NINE; }
        function variableCharacterRole(c) {
            if (c === DOT) {
                return TSCharRole.UnsignificantStarterChar;
            }
            if ((upper(c) >= A && upper(c) <= Z)) {
                return TSCharRole.StarterChar;
            }
            if (isVariableNum(c) || c === DOT || c === UNDERSCORE) {
                return TSCharRole.InternalChar;
            }
            return TSCharRole.ForbiddenChar;
        }
        function find_constant(c) {
            for (let cst of constants) {
                if (cst.name[0] === c)
                    return cst;
            }
            return undefined;
        }
        function store_variable(m, cm, variable, context) {
            const path = variable.split('.');
            for (let component of path) {
                if (component.length > 0) {
                    let cset = m.get(component);
                    if (!(0, commons_1.$ok)(cset)) {
                        cset = new Set();
                        m.set(component, cset);
                    }
                    cset.add(context);
                    let sset = cm.get(context);
                    if (!(0, commons_1.$ok)(sset)) {
                        sset = new Set();
                        cm.set(context, sset);
                    }
                    sset.add(component);
                    break;
                }
            }
        }
        function tohex(v) { return (0, commons_1.$isnumber)(v) ? `0x${v.toString(16)}` : `'${v}'`; }
        function tochar(c) { return String.fromCharCode(c); }
        if (debug) {
            (0, utils_1.$logterm)(`&0&x - &w source length&x:         &o${len}&0`);
            (0, utils_1.$logterm)(`&0&x - &w starting mark length&x:  &o${smlen}&0`);
            (0, utils_1.$logterm)(`&0&x - &w ending mark length&x:    &o${emlen}&0`);
            (0, utils_1.$logterm)(`&0&x - &w separator length&x:      &o${sepalen}&0`);
        }
        if (!len || !smlen || !sepalen || !emlen || smlen + emlen + sepalen > len) {
            throw new tserrors_1.TSError('template string to small', { startingMark: sm, endingMark: em, separator: sepa, source: source });
        }
        FUSION_SPECIAL_CHARS.forEach(c => {
            if (sm.includes(c) || em.includes(c) || sepa.includes(c)) {
                throw new tserrors_1.TSError(`starting mark, ending mark and seperator must not containe ${tohex(c)} ('${tochar(c)}') character`, { startingMark: sm, endingMark: em, separator: sepa, character: c });
            }
        });
        if (sm[0] === COMMA) {
            throw new tserrors_1.TSError('first character of starting mark cannot be a comma', { startingMark: sm, endingMark: em, separator: sepa });
        }
        if (em[0] === LPAR || em[0] === COMMA) {
            throw new tserrors_1.TSError('first character of ending mark cannot be a left parenthesis or a comma', { startingMark: sm, endingMark: em, separator: sepa });
        }
        if (sepa[0] === LPAR || sepa[0] === RPAR || sepa[0] === COMMA) {
            throw new tserrors_1.TSError('first character of separator cannot be a parenthesis or a comma', { startingMark: sm, endingMark: em, separator: sepa });
        }
        if (sm[0] === em[0] || sm[0] === sepa[0] || em[0] === sepa[0]) {
            throw new tserrors_1.TSError('first character of starting mark, ending mark and seperator must be distinct', { startingMark: sm, endingMark: em, separator: sepa });
        }
        let State;
        (function (State) {
            State["Text"] = "Text";
            State["OpenVariable"] = "open-var";
            State["PreVariableSpace"] = "pre-var-space";
            State["StartDecodingVariable"] = "start-decoding-var";
            State["DecodingVariable"] = "decoding-var";
            State["DecodeEndingMark"] = "decoding-end-mark";
            State["PostVariableSpace"] = "post-var-space";
            State["DecodingClosingVariable"] = "decoding-closing-var";
            State["DecodingSeparator"] = "decoding-separator";
            State["DecodingParameters"] = "decoding-parameters";
            State["NumberParameter"] = "decoding-number-param";
            State["ConstantParameter"] = "decoding-constant-param";
            State["StringParameter"] = "decoding-string-param";
            State["StringParamBackslash"] = "decoding-string-param-backslash";
            State["StringParamUnicode"] = "decoding-string-param-unicode";
            State["AfterParameter"] = "decoding-after-param";
            State["EndOfParameters"] = "decoding-closing-params";
        })(State || (State = {}));
        ;
        let state = State.Text;
        let tokenPos = 0;
        let variable = '';
        let t;
        let parameters = [];
        let currentParameter = undefined;
        let currentConstant = undefined;
        let stringEnder = QUOTE;
        let characterRole;
        const root = tsfusionnode_1.TSFusionTreeNode.root();
        let current = root;
        let last = 0;
        let beginText = 0;
        let fusionContext = tsfusionnode_1.TSFusionContextType.Local;
        let i = 0;
        let comp;
        let unicodeValue = 0;
        while (i < len) {
            const c = source[i];
            if (debug) {
                (0, utils_1.$logterm)(`&0&g${i.fpad(8, '_')}&w: &x&o${tohex(c)}&x (&b'&c${tochar(c)}&b'&x), state:&o${state}&x, current: &p${current.label}&x/&y${current.type}&0`);
            }
            switch (state) {
                case State.Text:
                    if (c === sm[0] || c === em[0]) {
                        last = i;
                        tokenPos = 0;
                        state = c === sm[0] ? State.OpenVariable : State.DecodeEndingMark;
                        i--;
                    }
                    break;
                case State.OpenVariable:
                    comp = sm[tokenPos++];
                    if (c !== comp) {
                        tokenPos = 0;
                        state = State.Text;
                        i = last;
                    }
                    else if (tokenPos === smlen) {
                        tokenPos = 0;
                        state = State.PreVariableSpace;
                        if (beginText >= 0 && beginText < last) {
                            current.pushData(source.slice(beginText, last));
                        }
                        beginText = -1;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                    }
                    break;
                case State.PreVariableSpace:
                    if (isVariableSpace(c)) {
                        break;
                    }
                    const ct = TSFusionContextTypes[c];
                    if ((0, commons_1.$ok)(ct)) {
                        fusionContext = ct;
                        tokenPos = i + 1;
                        state = State.StartDecodingVariable;
                    }
                    else {
                        characterRole = variableCharacterRole(c);
                        if (characterRole === TSCharRole.UnsignificantStarterChar || characterRole === TSCharRole.StarterChar) {
                            tokenPos = i;
                            state = characterRole === TSCharRole.UnsignificantStarterChar ? State.StartDecodingVariable : State.DecodingVariable;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed var : found forbidden initial variable character ${tohex(c)} ('${tochar(c)}') at position ${i}'.`, {
                                source: source,
                                character: c,
                                position: i,
                                state: state
                            });
                        }
                    }
                    break;
                case State.StartDecodingVariable:
                    characterRole = variableCharacterRole(c);
                    if (characterRole !== TSCharRole.UnsignificantStarterChar) {
                        if (characterRole === TSCharRole.StarterChar) {
                            state = State.DecodingVariable;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                                source: source,
                                character: c,
                                position: i,
                                state: state
                            });
                        }
                    }
                    break;
                case State.DecodingVariable:
                    t = TSFusionNodeTypes[c];
                    if ((0, commons_1.$ok)(t) || isVariableSpace(c)) {
                        variable = _data2ascii(source.slice(tokenPos, i));
                        current = current.pushVariable(variable, (0, commons_1.$ok)(t) ? t : tsfusionnode_1.TSFusionNodeType.Variable, fusionContext);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.PostVariableSpace;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        tokenPos = 0;
                    }
                    else if (c === LPAR) {
                        variable = _data2ascii(source.slice(tokenPos, i));
                        state = State.DecodingParameters;
                        parameters = [];
                        currentParameter = undefined;
                        tokenPos = 0;
                    }
                    else if (c === em[0]) {
                        variable = _data2ascii(source.slice(tokenPos, i));
                        current = current.pushVariable(variable, tsfusionnode_1.TSFusionNodeType.Variable, fusionContext);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.DecodingClosingVariable;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        i--;
                        tokenPos = 0;
                    }
                    else if (variableCharacterRole(c) === TSCharRole.ForbiddenChar) {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                            source: source,
                            character: c,
                            position: i,
                            state: state
                        });
                    }
                    break;
                case State.DecodingParameters:
                    if (c === RPAR) {
                        state = State.EndOfParameters;
                    }
                    else if (c === COMMA) {
                        parameters.push(currentParameter);
                        currentParameter = undefined;
                    }
                    else if (c === QUOTE || c === DOUBLE_QUOTE || c === PIPE) {
                        stringEnder = c;
                        currentParameter = new tsdata_1.TSData();
                        state = State.StringParameter;
                    }
                    else if (isVariableNum(c) || c === MINUS || c === PLUS || c === DOT) {
                        tokenPos = i;
                        state = State.NumberParameter;
                    }
                    else {
                        currentConstant = find_constant(c);
                        if ((0, commons_1.$defined)(currentConstant)) {
                            tokenPos = 1;
                            state = State.ConstantParameter;
                        }
                        else if (!isVariableSpace(c)) {
                            throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                                source: source,
                                character: c,
                                position: i,
                                neededChars: [TAB, SPACE],
                                state: state
                            });
                        }
                    }
                    break;
                case State.AfterParameter:
                    if (c === RPAR) {
                        parameters.push(currentParameter);
                        state = State.EndOfParameters;
                    }
                    else if (c === COMMA) {
                        parameters.push(currentParameter);
                        currentParameter = undefined;
                        state = State.DecodingParameters;
                    }
                    else if (!isVariableSpace(c)) {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [TAB, SPACE],
                            state: state
                        });
                    }
                    break;
                case State.ConstantParameter:
                    if (c === currentConstant.name[tokenPos]) {
                        tokenPos++;
                        if (tokenPos === currentConstant.len) {
                            currentParameter = currentConstant.value;
                            state = State.AfterParameter;
                        }
                    }
                    else {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (wrong parameter constant).`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [currentConstant.name[tokenPos]],
                            state: state
                        });
                    }
                    break;
                case State.NumberParameter:
                    if (isVariableSpace(c) || c === RPAR || c === COMMA) {
                        const num = Number(paramCharset === null || paramCharset === void 0 ? void 0 : paramCharset.stringFromData(source.slice(tokenPos, i)));
                        currentParameter = (0, commons_1.$isnumber)(num) ? num : 0;
                        state = State.AfterParameter;
                        i--;
                    }
                    else if (!isVariableNum(c) && c !== DOT) {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (parameter is not numeric).`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [ZERO, ZERO + 1, ZERO + 2, ZERO + 3, ZERO + 4, ZERO + 5, ZERO + 6, ZERO + 7, ZERO + 8, NINE],
                            state: state
                        });
                    }
                    break;
                case State.StringParameter:
                    if (c === stringEnder) {
                        currentParameter = currentParameter.toString(paramCharset);
                        if (stringEnder === PIPE) {
                            try {
                                currentParameter = JSON.parse(currentParameter);
                            }
                            catch (e) {
                                throw new tserrors_1.TSError(`Malformed JSON structure terminating at position ${i}. Impossible to parse. See 'parsingError' in complement infos.`, {
                                    source: source,
                                    position: i,
                                    state: state,
                                    parsingError: e
                                });
                            }
                        }
                        state = State.AfterParameter;
                    }
                    else if (c === BACKSLASH) {
                        state = State.StringParamBackslash;
                    }
                    else {
                        currentParameter.appendByte(c);
                    }
                    break;
                case State.StringParamBackslash:
                    state = State.StringParameter;
                    switch (c) {
                        case bu:
                            currentParameter.appendByte(BS);
                            break;
                        case fu:
                            currentParameter.appendByte(FF);
                            break;
                        case nu:
                            currentParameter.appendByte(LF);
                            break;
                        case ru:
                            currentParameter.appendByte(CR);
                            break;
                        case tu:
                            currentParameter.appendByte(TAB);
                            break;
                        case Uu:
                        case uu:
                            tokenPos = 0;
                            unicodeValue = 0;
                            state = State.StringParamUnicode;
                            break;
                        default:
                            if (c >= SPACE && c <= 0x7F) {
                                currentParameter.appendByte(c);
                            }
                            else {
                                const neededChars = [];
                                for (let nc = 32; nc < 128; nc++) {
                                    neededChars.push(nc);
                                }
                                throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Need to have an ASCII character after backslash in string parameter interpretation).`, {
                                    source: source,
                                    character: c,
                                    position: i,
                                    neededChars: neededChars,
                                    state: state
                                });
                            }
                    }
                    break;
                case State.StringParamUnicode:
                    const hexa = c.hexaValue();
                    if (hexa >= 0) {
                        unicodeValue = (unicodeValue << 4) | hexa;
                        tokenPos++;
                        if (tokenPos === 4) {
                            const bytes = paramCharset.bytesFromString(String.fromCharCode(unicodeValue));
                            currentParameter.appendBytes(bytes);
                            state = State.StringParameter;
                        }
                    }
                    else {
                        throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Need hexa character in string parameter unicode \\u sequence).`, {
                            source: source,
                            character: c,
                            position: i,
                            state: state
                        });
                    }
                    break;
                case State.EndOfParameters:
                    t = TSFusionNodeTypes[c];
                    if ((0, commons_1.$ok)(t) || isVariableSpace(c)) {
                        current = current.pushVariable(variable, (0, commons_1.$ok)(t) ? t : tsfusionnode_1.TSFusionNodeType.Variable, fusionContext, parameters);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.PostVariableSpace;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        tokenPos = 0;
                    }
                    else if (c === em[0]) {
                        current = current.pushVariable(variable, tsfusionnode_1.TSFusionNodeType.Variable, fusionContext, parameters);
                        store_variable(this._variables, this._contextsVariables, variable, fusionContext);
                        state = State.DecodingClosingVariable;
                        fusionContext = tsfusionnode_1.TSFusionContextType.Local;
                        i--;
                        tokenPos = 0;
                    }
                    else {
                        throw new tserrors_1.TSError(`Malformed var : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i}.`, {
                            source: source,
                            character: c,
                            position: i,
                            state: state
                        });
                    }
                    break;
                case State.PostVariableSpace:
                    if (c === em[0]) {
                        if (current.isContainerVariable) {
                            throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Need to have a separator for a container variable).`, {
                                source: source,
                                character: c,
                                position: i,
                                neededChars: [sm[0]],
                                state: state
                            });
                        }
                        state = State.DecodingClosingVariable;
                        i--;
                        tokenPos = 0;
                        break;
                    }
                    else if (c === sepa[0]) {
                        if (!current.isContainerVariable) {
                            throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (Separator forbidden for non container variable).`, {
                                source: source,
                                character: c,
                                position: i,
                                neededChars: [em[0]],
                                state: state
                            });
                        }
                        state = State.DecodingSeparator;
                        i--;
                        tokenPos = 0;
                        break;
                    }
                    else if (!isVariableSpace(c)) {
                        throw new tserrors_1.TSError(`Malformed template structure : found forbidden character ${tohex(c)} ('${tochar(c)}') at position ${i} (needed separator or ending mark start').`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [sepa[0], em[0]],
                            state: state
                        });
                    }
                    break;
                case State.DecodingSeparator:
                    comp = sepa[tokenPos++];
                    if (c !== comp) {
                        throw new tserrors_1.TSError(`Malformed template structure on variable separator. Unatended character  ${tohex(c)} ('${tochar(c)}') at position ${i} while decoding separator.`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [comp],
                            state: state
                        });
                    }
                    else if (tokenPos === sepalen) {
                        beginText = i + 1;
                        state = State.Text;
                    }
                    break;
                case State.DecodingClosingVariable:
                    comp = em[tokenPos++];
                    if (c !== comp) {
                        throw new tserrors_1.TSError(`Malformed template structure on closing variable. Unatended character  ${tohex(c)} ('${tochar(c)}') at position ${i} while decoding ending mark.`, {
                            source: source,
                            character: c,
                            position: i,
                            neededChars: [comp],
                            state: state
                        });
                    }
                    else if (tokenPos === emlen) {
                        current = current.parent;
                        state = State.Text;
                        beginText = i + 1;
                    }
                    break;
                case State.DecodeEndingMark:
                    comp = em[tokenPos++];
                    if (c !== comp) {
                        tokenPos = 0;
                        state = State.Text;
                        i = last;
                    }
                    else if (tokenPos === emlen) {
                        const parent = current.parent;
                        if ((0, commons_1.$ok)(parent)) {
                            if (beginText >= 0 && beginText < last) {
                                current.pushData(source.slice(beginText, last));
                            }
                            current = parent;
                            state = State.Text;
                            beginText = i + 1;
                        }
                        else {
                            throw new tserrors_1.TSError(`Malformed template structure. We have found a supernumerary ending mark at position ${i}.`, {
                                source: source,
                                position: i - emlen + 1,
                                endingMark: em,
                                state: state
                            });
                        }
                    }
                    break;
            }
            i++;
        }
        if (debug) {
            (0, utils_1.$logterm)(`&0&g--STOP--&w: &rEOF&x, state:&o${state}&x, current: &p${current.label}&0`);
        }
        if ((0, commons_1.$ok)(current.parent)) {
            throw new tserrors_1.TSError(`Malformed template structure. Template is not properly closed.`, {
                source: source,
                state: state
            });
        }
        else if (state === State.Text || state === State.OpenVariable || state === State.DecodeEndingMark) {
            if (beginText >= 0 && beginText < i) {
                current.pushData(source.slice(beginText, i));
            }
        }
        else {
            throw new tserrors_1.TSError(`Malformed template structure. Template is not properly closed (automat in wrong state '${state}').`, {
                source: source,
                state: state
            });
        }
        this.startingMark = sm;
        this.endingMark = em;
        this.separator = sepa;
        this.source = new tsdata_1.TSData(source);
        this._treeRoot = root;
        this._capacity = (0, commons_1.$capacityForCount)(this.source.length * 1.25);
        if ((0, commons_1.$ok)(opts.procedures)) {
            this._procedures = opts.procedures;
        }
        this._globalFunctions = opts.globalFunctions;
        this._addStandardGlobalFunctions = !!opts.addStandardGlobalFunctions;
    }
    get capacity() { return this._capacity; }
    set capacity(n) {
        const c = (0, commons_1.$capacityForCount)(n);
        if (c > this._capacity) {
            this._capacity = (this._capacity + (0, commons_1.$capacityForCount)(n));
        }
    }
    variablesForType(type) {
        const a = this._contextsVariables.get(type);
        return (0, commons_1.$ok)(a) && a.size > 0 ? Array.from(a) : [];
    }
    isVariableOfType(s, type) {
        if (!this.isVariable(s)) {
            return false;
        }
        const a = this._contextsVariables.get(type);
        return (0, commons_1.$ok)(a) && a.has(s);
    }
    get variables() { return Array.from(this._variables.keys()); }
    get userVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.User); }
    get globalVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Global); }
    get systemVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.System); }
    get localVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Local); }
    get rootVariables() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Root); }
    get procedures() { return this.variablesForType(tsfusionnode_1.TSFusionContextType.Procedure); }
    isVariable(s) { return (0, commons_1.$ok)(this._variables.get(s)); }
    isUserVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.User); }
    isGlobalVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Global); }
    isSystemVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.System); }
    isLocalVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Local); }
    isRootVariable(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Root); }
    isProcedure(s) { return this.isVariableOfType(s, tsfusionnode_1.TSFusionContextType.Procedure); }
}
exports.TSFusionTemplate = TSFusionTemplate;
TSFusionTemplate.DefaultStartingMark = '{{';
TSFusionTemplate.DefaultSeparator = ':';
TSFusionTemplate.DefaultEndingMark = '}}';
class TSStringTemplate extends TSFusionTemplate {
    static fromString(source, opts) {
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        try {
            return new TSStringTemplate(source, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0\n&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(source, options) {
        super(TSStringTemplate.__stringCharset.dataFromString(source), options, TSStringTemplate.__stringCharset);
        this._outputString = undefined;
        this.fusionStringWithDataContext = this.fusionWithDataContext;
    }
    pushFusionData(data) { this._outputString += TSStringTemplate.__stringCharset.stringFromData(data); }
    pushFusionValue(value) { this._outputString += (0, commons_1.$string)(value); }
    fusionWithDataContext(data, globalContext, errors) {
        this._outputString = '';
        return this._treeRoot.fusion(this, data, {
            globalContext: globalContext,
            procedures: this._procedures,
            errors: errors,
            globalFunctions: this._globalFunctions,
            addStandardGlobalFunctions: this._addStandardGlobalFunctions
        }) ? this._outputString : null;
    }
}
exports.TSStringTemplate = TSStringTemplate;
TSStringTemplate.__stringCharset = tscharset_1.TSCharset.utf8Charset();
class TSGenericDataTemplate extends TSFusionTemplate {
    constructor() {
        super(...arguments);
        this._outputData = undefined;
    }
    pushFusionData(data) { var _a; (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendBytes(data); }
    fusionWithDataContext(data, globalContext, errors) {
        this._outputData = new tsdata_1.TSData(this._capacity);
        return this._treeRoot.fusion(this, data, {
            globalContext: globalContext,
            procedures: this._procedures,
            errors: errors,
            globalFunctions: this._globalFunctions,
            addStandardGlobalFunctions: this._addStandardGlobalFunctions
        }) ? this._outputData : null;
    }
}
exports.TSGenericDataTemplate = TSGenericDataTemplate;
class TSHTMLTemplate extends TSGenericDataTemplate {
    static fromHTMLData(src, opts) {
        const source = (0, data_1.$uint8ArrayFromDataLike)(src);
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        try {
            return new TSHTMLTemplate(source, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0\n&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(src, opts) {
        const [data, charset] = _parseHTML((0, data_1.$uint8ArrayFromDataLike)(src), opts);
        super(data, opts, charset);
        this._htmlCharset = charset;
    }
    pushFusionValue(value) {
        var _a;
        let s = "";
        if ((0, commons_1.$ismethod)(value, 'toHTML')) {
            s = value.toHTML();
        }
        else if ((0, commons_1.$isnumber)(value)) {
            s = '' + value;
        }
        else {
            s = (0, commons_1.$string)(value).toHTML();
        }
        if (s.length) {
            (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendBytes(_ascii2data(s));
        }
    }
    fusionStringWithDataContext(data, globalContext, errors) {
        const ret = this.fusionWithDataContext(data, globalContext, errors);
        return (0, commons_1.$ok)(ret) ? this._htmlCharset.stringFromData(ret) : null;
    }
}
exports.TSHTMLTemplate = TSHTMLTemplate;
class TSDataTemplate extends TSGenericDataTemplate {
    static fromData(src, encoding, opts) {
        const source = (0, data_1.$uint8ArrayFromDataLike)(src);
        if (!(0, commons_1.$length)(source)) {
            return null;
        }
        const charset = (0, tscharset_1.$charset)(encoding);
        if (TSFusionForbiddenEncodings.includes(charset.name)) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)(`&0\n&R&w Data Parsing cannot use encoding: '${charset.name}'&0`);
            }
            return null;
        }
        try {
            return new TSDataTemplate(source, charset, opts);
        }
        catch (e) {
            if (opts === null || opts === void 0 ? void 0 : opts.debugParsing) {
                (0, utils_1.$logterm)('&0&R&w Data Parsing did encounter error &0\n&c' + (0, utils_1.$inspect)(e) + '\n&o' + e.stack + "&0");
            }
            return null;
        }
    }
    constructor(source, charset, opts) {
        super(source, opts, charset);
        this._charset = charset;
    }
    get charset() { return this._charset; }
    pushFusionValue(value) {
        var _a;
        (_a = this._outputData) === null || _a === void 0 ? void 0 : _a.appendString((0, commons_1.$string)(value), this._charset);
    }
    fusionStringWithDataContext(data, globalContext, errors) {
        const ret = this.fusionWithDataContext(data, globalContext, errors);
        return (0, commons_1.$ok)(ret) ? this.charset.stringFromData(ret) : null;
    }
}
exports.TSDataTemplate = TSDataTemplate;
const ASCIICharset = tscharset_1.TSCharset.asciiCharset();
function _ascii2data(s) { return ASCIICharset.bytesFromString(s); }
function _data2ascii(d, start, end) { return (0, commons_1.$ok)(d) ? ASCIICharset.stringFromBytes(d, start, end) : ''; }
function _parseHTML(src, options) {
    const startingMark = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.startingMark) ? options.startingMark : TSFusionTemplate.DefaultStartingMark;
    const endingMark = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.endingMark) ? options.endingMark : TSFusionTemplate.DefaultEndingMark;
    const separator = (0, commons_1.$length)(options === null || options === void 0 ? void 0 : options.separator) ? options.separator : TSFusionTemplate.DefaultSeparator;
    const debug = options === null || options === void 0 ? void 0 : options.debugParsing;
    const standardCharset = tscharset_1.TSCharset.utf8Charset();
    let currentCharset = standardCharset;
    if (debug) {
        (0, utils_1.$logterm)(`&0\n&P&w _parseHTML() is in debug mode &0`);
        (0, utils_1.$logterm)(`&0&x - &c starting mark&x:        &p${startingMark}&0`);
        (0, utils_1.$logterm)(`&0&x - &c ending mark&x:          &p${endingMark}&0`);
        (0, utils_1.$logterm)(`&0&x - &c separator&x:            &p${separator}&0`);
    }
    const sm = _ascii2data(startingMark);
    const em = _ascii2data(endingMark);
    const sp = _ascii2data(separator);
    let State;
    (function (State) {
        State["Text"] = "text";
        State["Start"] = "start";
        State["CloserElement"] = "closer-element";
        State["Element"] = "element";
        State["Inside"] = "inside";
        State["Tag"] = "tag";
        State["AfterTag"] = "after-tag";
        State["AfterEqual"] = "after-equal";
        State["Quote"] = "quote";
        State["AfterQuote"] = "after-quote";
        State["Value"] = "value";
        State["Slash"] = "slash";
    })(State || (State = {}));
    ;
    let Element;
    (function (Element) {
        Element["Unknown"] = "";
        Element["Fusion"] = "FUSION";
        Element["EndFusion"] = "/FUSION";
        Element["Meta"] = "META";
    })(Element || (Element = {}));
    const assign = {
        'FUSION': ['TYPE', 'PATH', 'CONTEXT'],
        'META': ['CHARSET'],
        '/FUSION': []
    };
    const types = {
        'IF': [QUESTION_MARK],
        'TEST': [QUESTION_MARK],
        'QUESTION': [QUESTION_MARK],
        '?': [QUESTION_MARK],
        'IFNOT': [EXCLAMATION_MARK],
        'NOT': [EXCLAMATION_MARK],
        '!TEST': [QUESTION_MARK],
        '!': [EXCLAMATION_MARK],
        'ENUM': [SHARP],
        'ENUMERATION': [SHARP],
        'LIST': [SHARP],
        'ARRAY': [SHARP],
        '#': [SHARP],
        'VALUE': [],
        'PATH': [],
        'VAR': [],
        'VARIABLE': [],
        '': []
    };
    const contexts = {
        'GLOBAL': [DOLLAR],
        '$': [DOLLAR],
        'DICT': [DOLLAR],
        'DICTIONARY': [DOLLAR],
        'ROOT': [AT],
        'INITIAL': [AT],
        'TEMPLATE': [AT],
        '@': [AT],
        'SYS': [UNDERSCORE],
        'SYSTEM': [UNDERSCORE],
        '_': [UNDERSCORE],
        'PROC': [ASTERISK],
        'PROCEDURE': [ASTERISK],
        'FN': [ASTERISK],
        'FUNC': [ASTERISK],
        'FUNCTION': [ASTERISK],
        'CALL': [ASTERISK],
        '*': [ASTERISK],
        'USER': [PLUS],
        'USER-DEFINED': [PLUS],
        'PARAM': [PLUS],
        'PARAMETER': [PLUS],
        '+': [PLUS],
        'CURRENT': [DOT],
        'LOCAL': [DOT],
        '.': [DOT],
        '': [DOT]
    };
    const len = src.length;
    let target = new tsdata_1.TSData(len);
    let state = State.Text;
    let i = 0;
    let last = 0;
    let lastQuote = 0;
    let insideLast = 0;
    let tag = '';
    let element = Element.Unknown;
    let values = {};
    let isCloser = false;
    function upper(c) { return c & 0xdf; }
    function isNameStarterChar(c) { c = upper(c); return c >= A && c <= Z; }
    function isNameChar(c) { return c === DOT || c === COLON || c === UNDERSCORE || c === MINUS || (c >= ZERO && c <= NINE) || isNameStarterChar(c); }
    function isWhiteSpace(c) { return c === TAB || c === SPACE || c === LF || c === CR; }
    function isQuote(c) { return c === QUOTE || c === DOUBLE_QUOTE; }
    function name(start, end) { return _data2ascii(src, start, end).toUpperCase(); }
    function registerValue(start, end) {
        var _a;
        if ((_a = assign[element]) === null || _a === void 0 ? void 0 : _a.includes(tag)) {
            values[tag] = src.slice(start, end);
        }
        tag = '';
    }
    function reset() {
        state = State.Text;
        element = Element.Unknown;
        tag = '';
        last = 0;
        values = {};
        isCloser = false;
    }
    function back() {
        target.appendByte(LT);
        i = last;
        reset();
    }
    function endElement() {
        switch (element) {
            case Element.Fusion:
                const p = values['PATH'];
                if (!(0, commons_1.$length)(p)) {
                    back();
                    return;
                }
                const c = contexts[_data2ascii(values['CONTEXT']).toUpperCase()];
                const t = types[_data2ascii(values['TYPE']).toUpperCase()];
                target.appendBytes(sm);
                target.appendBytes(c);
                target.appendBytes(p);
                if (t.length) {
                    target.appendBytes(t);
                    target.appendBytes(sp);
                }
                else {
                    target.appendBytes(em);
                }
                reset();
                return;
            case Element.Meta:
                const cset = values['CHARSET'];
                if ((0, commons_1.$length)(cset)) {
                    const newCharset = tscharset_1.TSCharset.charset(_data2ascii(cset));
                    if ((0, commons_1.$ok)(newCharset) && newCharset !== standardCharset) {
                        currentCharset = newCharset;
                    }
                }
                target.appendBytes(src, last, i);
                target.appendByte(GT);
                break;
            case Element.EndFusion:
                target.appendBytes(em);
                reset();
                return;
            default:
                target.appendBytes(src, last, i);
                target.appendByte(GT);
                break;
        }
        reset();
    }
    while (i < len) {
        const c = src[i];
        if (debug) {
            (0, utils_1.$logterm)(`&0&j${i.fpad(8, '_')}&w: &x0x&p${c.toString(16)}&x (&b'&c${String.fromCharCode(c)}&b'&x), state:&p${state}&x, last:&y${last}&0`);
        }
        switch (state) {
            case State.Text:
                if (c === LT) {
                    state = State.Start;
                    last = i;
                }
                else {
                    target.appendByte(c);
                }
                break;
            case State.Start:
                if (c === SLASH) {
                    state = State.CloserElement;
                    isCloser = true;
                }
                else if (isNameStarterChar(c)) {
                    state = State.Element;
                }
                else {
                    back();
                }
                break;
            case State.CloserElement:
                if (isNameStarterChar(c)) {
                    state = State.Element;
                }
                else {
                    back();
                }
                break;
            case State.Element:
                if (c === GT || isWhiteSpace(c)) {
                    const elementName = name(last + 1, i).toUpperCase();
                    if (!(0, commons_1.$ok)(assign[elementName])) {
                        back();
                    }
                    else {
                        element = elementName;
                        if (c === GT) {
                            endElement();
                        }
                        else {
                            state = State.Inside;
                        }
                    }
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.Inside:
                if (c === GT) {
                    endElement();
                }
                else if (isCloser && !isWhiteSpace(c)) {
                    back();
                }
                else if (c === SLASH) {
                    state = State.Slash;
                }
                else if (isNameStarterChar(c)) {
                    insideLast = i;
                    state = State.Tag;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.Tag:
                if (c === EQ || isWhiteSpace(c)) {
                    tag = name(insideLast, i);
                    state = c === EQ ? State.AfterEqual : State.AfterTag;
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.AfterTag:
                if (c === EQ) {
                    state = State.AfterEqual;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.AfterEqual:
                if (isQuote(c)) {
                    insideLast = i + 1;
                    lastQuote = c;
                    state = State.Quote;
                }
                else if (isNameStarterChar(c)) {
                    insideLast = i;
                    state = State.Value;
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
            case State.Value:
                if (c === GT || c === SLASH || isWhiteSpace(c)) {
                    registerValue(insideLast, i);
                    if (c === GT) {
                        endElement();
                    }
                    else if (c === SLASH) {
                        state = State.Slash;
                    }
                    else {
                        state = State.Inside;
                    }
                }
                else if (!isNameChar(c)) {
                    back();
                }
                break;
            case State.Quote:
                if (c === lastQuote) {
                    registerValue(insideLast, i);
                    state = State.AfterQuote;
                }
                break;
            case State.AfterQuote:
                if (isWhiteSpace(c)) {
                    state = State.Inside;
                }
                else if (c === SLASH) {
                    state = State.Slash;
                }
                else if (c === GT) {
                    endElement();
                }
                else {
                    back();
                }
                break;
            case State.Slash:
                if (c === GT) {
                    endElement();
                }
                else if (!isWhiteSpace(c)) {
                    back();
                }
                break;
        }
        i++;
        if (debug) {
            (0, utils_1.$logterm)(`&0&o${i.fpad(8, '_')}&w: ----------, state:&p${state}&x, last:&y${last}&0`);
        }
    }
    if (state !== State.Text) {
        target.appendByte(LT);
        target.appendBytes(src, last, i);
        target.appendByte(GT);
    }
    if (debug) {
        (0, utils_1.$logterm)(`&0&j--STOP--&w: &rEOF&x, state:&p${state}&x, last:&p${last}&0`);
    }
    return [target, currentCharset];
}
//# sourceMappingURL=tsfusion.js.map