"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSFusionTreeNode = exports.TSFusionNodeType = exports.TSFusionContextType = void 0;
const array_1 = require("./array");
const commons_1 = require("./commons");
const compare_1 = require("./compare");
const data_1 = require("./data");
const date_1 = require("./date");
const number_1 = require("./number");
const strings_1 = require("./strings");
const tsdate_1 = require("./tsdate");
const tsdefaults_1 = require("./tsdefaults");
const tserrors_1 = require("./tserrors");
const tsobject_1 = require("./tsobject");
const utils_1 = require("./utils");
var TSFusionContextType;
(function (TSFusionContextType) {
    TSFusionContextType["Local"] = "local";
    TSFusionContextType["Global"] = "global";
    TSFusionContextType["Root"] = "root";
    TSFusionContextType["System"] = "system";
    TSFusionContextType["Procedure"] = "procedure";
    TSFusionContextType["User"] = "user";
})(TSFusionContextType = exports.TSFusionContextType || (exports.TSFusionContextType = {}));
;
var TSFusionNodeType;
(function (TSFusionNodeType) {
    TSFusionNodeType["Root"] = "root";
    TSFusionNodeType["Data"] = "data";
    TSFusionNodeType["Variable"] = "variable";
    TSFusionNodeType["Test"] = "test-variable";
    TSFusionNodeType["NegativeTest"] = "nagative-test-variable";
    TSFusionNodeType["Enumeration"] = "enumeration-variable";
})(TSFusionNodeType = exports.TSFusionNodeType || (exports.TSFusionNodeType = {}));
class TSFusionTreeNode {
    static root() { return new TSFusionTreeNode(TSFusionNodeType.Root); }
    constructor(type, value, parent) {
        this.contextType = TSFusionContextType.Local;
        this._localContext = {};
        this.type = type;
        if ((0, commons_1.$ok)(value)) {
            this.value = value;
        }
        if ((0, commons_1.$ok)(parent)) {
            this.parent = parent;
        }
    }
    get label() {
        switch (this.type) {
            case TSFusionNodeType.Root: return 'Root';
            case TSFusionNodeType.Data: return 'Data';
            default: return this.value;
        }
    }
    get isContainerVariable() {
        return this.type === TSFusionNodeType.Test || this.type === TSFusionNodeType.NegativeTest || this.type === TSFusionNodeType.Enumeration;
    }
    pushData(data) {
        if (this.type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError('Impossible to add subnode to a text FusionTreeNode', { text: this.value });
        }
        if (data.length) {
            if (!(0, commons_1.$ok)(this.nodes)) {
                this.nodes = [];
            }
            this.nodes.push(new TSFusionTreeNode(TSFusionNodeType.Data, data, this));
        }
    }
    pushVariable(name, type, contextType, parameters) {
        if (this.type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError('Impossible to add subnode to a text FusionTreeNode', { text: this.value });
        }
        if (!(0, commons_1.$length)(name)) {
            throw new tserrors_1.TSError(`Impossible to add an unamed fusion variable to ${this.label} variable.`);
        }
        if (type === TSFusionNodeType.Root || type === TSFusionNodeType.Data) {
            throw new tserrors_1.TSError(`Impossible to add fusion variable '${name}' as a ${type} variable`, { name: name, type: type });
        }
        if (!(0, commons_1.$ok)(this.nodes)) {
            this.nodes = [];
        }
        let node = new TSFusionTreeNode(type, name, this);
        if ((0, commons_1.$ok)(contextType)) {
            node.contextType = contextType;
        }
        if ((0, commons_1.$count)(parameters)) {
            node.parameters = parameters;
        }
        node._pathCache = name.split('.');
        this.nodes.push(node);
        return node;
    }
    fusion(template, data, options = {}) {
        if (this.type !== TSFusionNodeType.Root) {
            throw new tserrors_1.TSError('Impossible invoque fusion() method on non-root FusionNodeType', { label: this.label, type: this.type });
        }
        try {
            this._cleanLocalContexts();
            const globalContext = (0, commons_1.$ok)(options.globalContext) ? options.globalContext : {};
            function _addGlobalFunctions(functions) {
                if ((0, commons_1.$ok)(functions)) {
                    for (let [name, fn] of Object.entries(functions)) {
                        if (!(0, commons_1.$ok)(globalContext[name])) {
                            globalContext[name] = fn;
                        }
                    }
                }
            }
            _addGlobalFunctions(options.globalFunctions);
            if (options.addStandardGlobalFunctions) {
                [
                    commons_1.$value, commons_1.$valueornull, commons_1.$valueorundefine,
                    commons_1.$isstring, commons_1.$iswhitespace, commons_1.$isnumber, commons_1.$isint, commons_1.$isunsigned, commons_1.$isbool, commons_1.$isobject, commons_1.$isarray, commons_1.$isiterable, commons_1.$isdate, commons_1.$isemail, commons_1.$isurl, commons_1.$isuuid, commons_1.$isfunction,
                    commons_1.$int, commons_1.$intornull, commons_1.$email, commons_1.$url, commons_1.$UUID, commons_1.$isodate, commons_1.$unsignedornull, commons_1.$unsigned, commons_1.$toint, commons_1.$tounsigned, commons_1.$string, commons_1.$array, commons_1.$strings,
                    commons_1.$count, commons_1.$length, commons_1.$objectcount, commons_1.$lengthin,
                    commons_1.$jsonobj, commons_1.$json, commons_1.$keys,
                    compare_1.$numcompare, compare_1.$datecompare, compare_1.$bytescompare, compare_1.$compare, compare_1.$bytesequal, compare_1.$equal, compare_1.$setequal, compare_1.$unorderedEqual,
                    data_1.$bufferFromArrayBuffer, data_1.$bufferFromBytes, data_1.$uint8ArrayFromBytes, data_1.$uint8ArrayFromDataLike, data_1.$arrayBufferFromBytes, data_1.$decodeBase64, data_1.$encodeBase64,
                    date_1.$timeBetweenDates,
                    number_1.$div, number_1.$icast, number_1.$round, number_1.$fpad, number_1.$fpad2, number_1.$fpad3, number_1.$fpad4, number_1.$octets, number_1.$meters,
                    strings_1.$ascii, strings_1.$trim, strings_1.$rtrim, strings_1.$ltrim, strings_1.$left, strings_1.$right, strings_1.$lines, strings_1.$firstcap, strings_1.$capitalize, strings_1.$normspaces, strings_1.$HTML,
                    tsdate_1.$isleap, tsdate_1.$dayisvalid, tsdate_1.$timeisvalid, tsdate_1.$timestamp, tsdate_1.$secondFromTimestamp, tsdate_1.$minuteFromTimestamp, tsdate_1.$hourFromTimestamp, tsdate_1.$timestampWithoutTime, tsdate_1.$dayOfYear, tsdate_1.$dayOfWeekFromTimestamp, tsdate_1.$weekOfYear, tsdate_1.$insettimestamp,
                    tsdefaults_1.$tmp, tsdefaults_1.$locales, tsdefaults_1.$country, tsdefaults_1.$language, tsdefaults_1.$currency, tsdefaults_1.$default,
                    utils_1.$inspect, utils_1.$insp, utils_1.$term, utils_1.$termclean, utils_1.$stack
                ].forEach(fn => {
                    const name = fn.name.slice(1);
                    if (!(0, commons_1.$ok)(globalContext[name])) {
                        globalContext[name] = fn;
                    }
                });
                _addGlobalFunctions({
                    'abs': Math.abs,
                    'sign': Math.sign,
                    'max': Math.max,
                    'min': Math.min,
                    'cos': Math.cos,
                    'sin': Math.sin,
                    'tan': Math.tan,
                    'log': Math.log,
                    'exp': Math.exp,
                    'sqrt': Math.sqrt
                });
            }
            const defaults = tsdefaults_1.TSDefaults.defaults();
            this._fusion(template, globalContext, [data], [this._localContext], {
                zulu: tsdate_1.TSDate.zulu(),
                date: new tsdate_1.TSDate(),
                language: defaults.language(),
                languages: defaults.managedLanguages(),
                currency: defaults.currency(),
                locales: defaults.locales(),
                index: 0,
                position: 1,
                remaining: 0,
                count: 1,
                browser: (0, utils_1.$inbrowser)(),
                log: (data) => { var _a; (_a = options.errors) === null || _a === void 0 ? void 0 : _a.push(`USERLOG:${data}`); return ''; }
            }, (0, commons_1.$value)(options.procedures, {}), (0, commons_1.$ok)(options.errors) ? options.errors : undefined);
            return true;
        }
        catch (e) {
            if ((0, commons_1.$ok)(options.errors)) {
                options.errors.push('Fusion did encounter error:');
                (0, strings_1.$lines)((0, utils_1.$inspect)(e)).forEach(l => { var _a; return (_a = options.errors) === null || _a === void 0 ? void 0 : _a.push(l); });
                options.errors.push('STACK:');
                (0, strings_1.$lines)(e.stack).forEach(l => { var _a; return (_a = options.errors) === null || _a === void 0 ? void 0 : _a.push(l); });
            }
        }
        return false;
    }
    _cleanLocalContexts() {
        var _a;
        this._localContext = {};
        (_a = this.nodes) === null || _a === void 0 ? void 0 : _a.forEach(n => n._cleanLocalContexts());
    }
    _fusion(template, globalContext, stack, localStack, systemContext, procedures, errors) {
        var _a;
        const debug = (0, commons_1.$ok)(errors);
        const count = (0, commons_1.$count)(this.nodes);
        for (let i = 0; i < count; i++) {
            const node = this.nodes[i];
            if (node.type === TSFusionNodeType.Data) {
                template.pushFusionData(node.value);
            }
            else {
                let v = undefined;
                if (node.contextType === TSFusionContextType.Procedure) {
                    if (((_a = node._pathCache) === null || _a === void 0 ? void 0 : _a.length) === 1 && (0, commons_1.$length)(node.value) > 0) {
                        const proc = procedures[node.value];
                        if ((0, commons_1.$isfunction)(proc)) {
                            try {
                                v = proc(stack.last(), stack.first(), this._localContext, globalContext, Object.assign({}, systemContext));
                            }
                            catch (e) {
                                if (debug) {
                                    errors.push(`!ERROR!: Procedure ${node.value}() execution did fail`);
                                }
                                v = undefined;
                            }
                        }
                        else if (debug) {
                            errors.push(`!ERROR!: Procedure ${node.value}() does not exist`);
                        }
                    }
                    else if (debug) {
                        errors.push(`!ERROR!: Malformed procedure name ${node.value}() (undefined, empty or with '.' character in it)`);
                    }
                }
                else {
                    let target = [];
                    let dataType = 'unknown target';
                    const hasRedefinedLocalItems = (0, commons_1.$objectcount)(localStack.last()) > 0;
                    switch (node.contextType) {
                        case TSFusionContextType.Root:
                            target = [stack.first()];
                            dataType = 'root';
                            break;
                        case TSFusionContextType.Local:
                            target = hasRedefinedLocalItems ? localStack : stack;
                            dataType = 'local context';
                            break;
                        case TSFusionContextType.Global:
                            target = [globalContext];
                            dataType = 'global context';
                            break;
                        case TSFusionContextType.User:
                            target = [globalContext];
                            dataType = 'user global context';
                            break;
                        case TSFusionContextType.System:
                            target = [systemContext];
                            dataType = 'system context';
                            break;
                    }
                    const parameters = (0, commons_1.$count)(node.parameters) ? (0, array_1.$map)(node.parameters, p => p === template ? stack.last() : p) : undefined;
                    v = _valueForKeyPath(target, node._pathCache, dataType, false, parameters, errors);
                    if (node.contextType === TSFusionContextType.Local && !(0, commons_1.$ok)(v) && hasRedefinedLocalItems) {
                        v = _valueForKeyPath(stack, node._pathCache, 'target object', true, parameters, errors);
                    }
                }
                switch (node.type) {
                    case TSFusionNodeType.Variable:
                        if ((0, commons_1.$ok)(v)) {
                            template.pushFusionValue(v);
                        }
                        else {
                            errors === null || errors === void 0 ? void 0 : errors.push(`WARNING: value returned for '${node.value}' was undefined or null and should have result in data to merge in fusion.`);
                        }
                        break;
                    case TSFusionNodeType.Test:
                        if (v) {
                            node._fusion(template, globalContext, stack, localStack, systemContext, procedures, errors);
                        }
                        break;
                    case TSFusionNodeType.NegativeTest:
                        if (!v) {
                            node._fusion(template, globalContext, stack, localStack, systemContext, procedures, errors);
                        }
                        break;
                    case TSFusionNodeType.Enumeration:
                        if ((0, commons_1.$ok)(v)) {
                            const a = (0, tsobject_1.$iscollection)(v) ? v.getItems() : ((0, commons_1.$ismethod)(v, 'fusionEnumeration') ? v.fusionEnumeration() : [v]);
                            const acount = (0, commons_1.$length)(a);
                            for (let j = 0; j < acount; j++) {
                                node._fusion(template, globalContext, [...stack, a[j]], [...localStack, a[j]._localContext], Object.assign(Object.assign({}, systemContext), { count: acount, index: j, position: j + 1, remaining: acount - j - 1 }), procedures, errors);
                            }
                        }
                        break;
                    default: break;
                }
            }
        }
    }
}
exports.TSFusionTreeNode = TSFusionTreeNode;
function _valueForKeyPath(stack, keyPath = [], dataType, unknownAsError, parameters, errors) {
    const scount = stack.length;
    const kcount = keyPath.length;
    if (!kcount) {
        errors === null || errors === void 0 ? void 0 : errors.push('ERROR: empty key path');
        return undefined;
    }
    let i = 0;
    for (; i < kcount && keyPath[i].length === 0; i++) { }
    if (i >= scount) {
        errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: key path '${keyPath.join('.')}' did underflow the stack.`);
        return undefined;
    }
    let v = stack[scount - i - 1];
    const pcount = (0, commons_1.$count)(parameters);
    for (; i < kcount && (0, commons_1.$ok)(v); i++) {
        const m = keyPath[i];
        const shouldHandleParameters = pcount > 0 && i + 1 === kcount;
        if (!m.length) {
            errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: key path '${keyPath.join('.')}' contains internal '..'.`);
            return undefined;
        }
        if (!shouldHandleParameters && m === 'self') {
            continue;
        }
        if ((0, commons_1.$isobject)(v) && !(m in v)) {
            errors === null || errors === void 0 ? void 0 : errors.push(`${unknownAsError ? '!ERROR!' : 'WARNING'}: unknown method or preperty '${m}' in ${dataType}.`);
            return undefined;
        }
        let res = v[m];
        if ((0, commons_1.$isfunction)(res)) {
            if (res.length > 0 && !shouldHandleParameters) {
                errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: method ${m}() is not an sinple accessor in ${dataType}.`);
                return undefined;
            }
            try {
                v = shouldHandleParameters ? res.apply(v, parameters) : res.call(v);
            }
            catch (e) {
                errors === null || errors === void 0 ? void 0 : errors.push(`!ERROR!: method ${m}() execution did fail on ${dataType}`);
                return undefined;
            }
        }
        else if (!shouldHandleParameters) {
            v = res;
        }
        else {
            errors === null || errors === void 0 ? void 0 : errors.push(`${unknownAsError ? '!ERROR!' : 'WARNING'}: unknown method '${m}()' in ${dataType}.`);
            return undefined;
        }
        if (typeof v === 'number' && isNaN(v)) {
            errors === null || errors === void 0 ? void 0 : errors.push(`WARNING: key path '${keyPath.join('.')}' would return NaN which was transformed to null.`);
            return null;
        }
    }
    return v;
}
Set.prototype.fusionEnumeration = function fusionEnumeration() { return Array.from(this); };
Map.prototype.fusionEnumeration = function fusionEnumeration() {
    const ret = [];
    this.forEach((value, key) => ret.push({ key: key, value: value }));
    return ret;
};
//# sourceMappingURL=tsfusionnode.js.map